#!/usr/bin/python3

import sys
import os
import subprocess
import math
import json

# For matrix questions, the column headers are given as a string
# argument to the QuestionMatrix environment in Latex. Which separator
# string (often just a char) is used to separate the headers? Change
# this if you need headers containing the current separator.
matrixQuestionColumnHeaderSeparator = ';'

# Place of resulting files (note: current dir should be designated "."
# (not empty string)):
nameOfResultDir = "DigitalEksamenMCQData"

nameOfJSONManifestFile = "manifest.json"

# Utility function:

## For input string s, the function returns a list of the (pairs of)
## indices in s of the outermost matching pairs of openchar and
## closechar. Thus balparscontents("xx{{xxx}}xx{}x", '{', '}') should
## return [(2,8),(11,12)].
def balparscontents(s, openchar, closechar):
    indexlist = []
    oldNoOfOpen = 0
    i = 0
    for c in s:
        if c == openchar:
            newNoOfOpen = oldNoOfOpen + 1
        elif c == closechar:
            newNoOfOpen = oldNoOfOpen - 1
        else:
            newNoOfOpen = oldNoOfOpen
        if newNoOfOpen < 0:
            print("Error: unbalanced expression!")
        if (newNoOfOpen,oldNoOfOpen) == (1,0):
            lastStartIndex = i # start index found
        if (newNoOfOpen,oldNoOfOpen) == (0,1): # matching end index found
            indexlist.append((lastStartIndex,i))
        oldNoOfOpen = newNoOfOpen
        i += 1
    if oldNoOfOpen > 0:
        print("Error: unbalanced expression!")
    return [s[i+1:j] for (i,j) in indexlist]

inputfilename = sys.argv[1]

# get input filename without extension:
inputfilebasename = os.path.splitext(inputfilename)[0]

# run pdflatex on the input tex file, making the preview package
# active (and with minimum messages)
print("Running pdflatex with preview package active.")
subprocess.call(r"pdflatex -interaction=batchmode '\PassOptionsToPackage{active}{preview}\input{%s.tex}'" % inputfilebasename, shell=True)

# input file as a string:
input = open(sys.argv[1],'r').read()

# store path to input file in base dir before moving to result dir
basedir = os.getcwd()
absinputfilebasename = basedir + "/" + inputfilebasename

# if necessary, create result dir
if not os.path.exists(nameOfResultDir):
  print(f"Making directory {nameOfResultDir}")
  subprocess.call(f"mkdir {nameOfResultDir}", shell=True)
# Move to designated result dir, make it empty
os.chdir(nameOfResultDir)
print(f"Removing any old files in {nameOfResultDir}")
subprocess.call(f"rm -f *.pdf *.png {nameOfJSONManifestFile}", shell=True)

# get rid of any questions parked below "\end{document} in input"
input = input.split(r"\end{document}",maxsplit=1)[0]

# chop input at beginnings of Question and Answer environments, giving
# a list with an entry for each question, the entry being a list of
# the question and its answers.
cuts = [q.split(r"\begin{Answer") for q in input.split(r"\begin{Question")[1:]]

# Next comes the main part: a double for-loop over cuts which
# generates a nested data structure (dicts and lists) which json.dump
# then outputs.

questions = []

qcounter = 1
acounter = 1
filebasenames = [] # for later accessing all picture files created

for qpartslist in cuts:
    acounter = 1
    answersdata = []
    # Note: the filenames should be generated in same order as the pdf
    # pages, hence the Qfilebasename should be generated first.
    Qfilebasename = f"Q{qcounter:03}"
    filebasenames.append(Qfilebasename)
    # Process the answers in qpartslist
    for a in qpartslist[1:]:
        Afilebasename = f"Q{qcounter:03}-A{acounter:03}"
        filebasenames.append(Afilebasename)
        answersdata.append((Afilebasename,a.startswith("T")))
        acounter += 1
    # Process the question in qpartslist
    # Is the question a QuestionMultipleAnswers latex environment?
    qfirstlineremainingpart = qpartslist[0].split("\n", maxsplit=1)[0]
    if qfirstlineremainingpart.startswith("MultipleAnswers}"):
        # From rest of qfirstlineending, extract contents of second
        # "{}" pair (i.e., get second parameter to latex
        # QuestionMultipleAnswers environment).
        questionParameterString = qfirstlineremainingpart[16:]
        columnHeadingsString = balparscontents(questionParameterString,
                                               '{','}')[1]
        # Here, we assume that matrix questions have exactly two
        # columns, first of which stands for correct
        h1,h2 = columnHeadingsString.split(matrixQuestionColumnHeaderSeparator)
        answers = [{"content": "{{%s.png}}" % Afilebasename,
                    "correctColumn": 0 if correct else 1}
                   for (Afilebasename,correct) in answersdata]
        question = {"type": "matrix", "content": "{{%s.png}}" % Qfilebasename,
                    "columns": [h1,h2], "answers": answers}
        
    else: # It is a plain Question latex environment (single correct answer)
        answers = [{"content": "{{%s.png}}" % Afilebasename, "correct": correct}
                   for (Afilebasename,correct) in answersdata]
        question = {"type": "basic", "content": "{{%s.png}}" % Qfilebasename,
                    "answers": answers}
    questions.append(question)
    qcounter += 1

# Create json output:
print(f"Writing json output to {nameOfJSONManifestFile}")
with open(nameOfJSONManifestFile,'w') as f:
    json.dump({"questions": questions}, f, indent=2)

print("Sanity checking")
# find number of pages in pdf file
pdfNoOfPages = int(subprocess.Popen(f"qpdf --show-npages {absinputfilebasename}.pdf", stdout=subprocess.PIPE, shell=True).communicate()[0])
noOfFileNames = len(filebasenames)
# sanity check:
print(f"There are {pdfNoOfPages} pages in pdf file")
print(f"There are {noOfFileNames} filenames in tex file")
if pdfNoOfPages != noOfFileNames:
  print(
"""These do not match! Exiting
""")
  sys.exit()
else:
  print("OK")

# For the --split-pages option in qpdf, a "%d" in the target file name
# is converted into 0-padded counter string of the necessary fixed
# width (as dictated by the total number of pages) when creating file
# names for the new files (one new file for each page). Hence, we need
# to know that width:
noOfDecimalDigits = int(math.ceil(math.log10(pdfNoOfPages + 0.5)))

print("Splitting pdf file")
# split pdf file into single page files:
subprocess.call(f"qpdf --split-pages {absinputfilebasename}.pdf pg_%d.pdf", shell=True)

print("Renaming resulting pdf files")
# rename all of these:
counter = 1
for filename in filebasenames:
  subprocess.call(f"mv pg_{counter:0{noOfDecimalDigits}}.pdf {filename}.pdf", shell=True)
  counter += 1

print("Making corresponding png files")
# convert all of these to png versions:
#(density = x gives scaling by factor x/72 while keeping quality fair)
for filename in filebasenames:
  subprocess.call(f"pdftoppm -png -r 144 -singlefile {filename}.pdf {filename}", shell=True)
  
print("Cleanup: Removing generated pdf files")
subprocess.call("rm *.pdf", shell=True)

# run pdflatex again on the input tex file, without the preview
# package active, in order to restore the look of the pdf of the main
# file.
print("Cleanup: Running pdflatex without preview package active (and with minimum messages).")
os.chdir(basedir)
subprocess.call(r"pdflatex -interaction=batchmode '\input{%s.tex}'" % inputfilebasename, shell=True)

print()
print(f"File {inputfilename} was converted to input for Digital Eksamen MCQ.")
print()
print(f"The resulting data is in directory {nameOfResultDir}.")
print()
print(f"To load it, use the browser extension Digital Eksamen Enhancer which ")
print(f"can be found at https://github.com/birjj/digitaleksamen-enhancer.")
print()
print(f"The place to load it is https://digitaleksamen.sdu.dk/, at the menu")
print(f"item Create (in Danish: Opret)")
